unit U_LatLongDistance;
 {Copyright   2004, Gary Darby,  www.DelphiForFun.org
 This program may be used or modified for any non-commercial purpose
 so long as this original notice remains in place.
 All other rights are reserved
 }

{Estimate surface distance between two points on the earth's surface defined
by Latitude and Longitude.  Two methods implemented, spherical earth assumption
and estimate considering elliptical earth}


interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, shellapi, Buttons;

type
  TForm1 = class(TForm)
    Label1: TLabel;
    Loc1Lat: TEdit;
    Label2: TLabel;
    Memo1: TMemo;
    UnitsGrp: TRadioGroup;
    SphereBtn: TButton;
    Loc1Lon: TEdit;
    Label3: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Loc2lat: TEdit;
    Label6: TLabel;
    Loc2Lon: TEdit;
    ResultsLbl: TLabel;
    StaticText1: TStaticText;
    EllispeBtn: TBitBtn;
    procedure SphereBtnClick(Sender: TObject);
    procedure StaticText1Click(Sender: TObject);
    procedure EllispeBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation
uses math;


{$R *.DFM}

var deg:char=chr(176); {''}
    minmark:char=chr(180);
    ERAD:extended = 6378.135; {Earth's radius in km a equator}
    FLATTENING:extended = 1.0/298.257223563; {Fractional reduction of radius to poles}
    unitslbl:array[0..2] of string=('Miles','Kilometers','Nautical miles');




 {********* Deg2rad ***********}
function deg2rad(deg:extended):extended;
{Degrees to Radians}
begin result:=deg * PI / 180.0; end;


{*********** Rad2Deg **********}
function rad2deg(rad:extended):extended;
{radians to degrees}
begin result:=rad / PI * 180.0;  end;

{**************** StrToAngle ************}
function StrtoAngle(const s:string; var angle:extended):boolean;
{Convert string representation of an angle to a numberic value}
           function convertfield(xs:string;var v:extended; msg:string):boolean;
           var errcode:integer;
           begin
             val(xs,v,errcode);
             if errcode<>0 then
             begin
               showmessage('Invalid character in '+msg+' field in string '+ s);
               result:=false;
             end
             else result:=true;
           end;
 var
   n,fn, sign:integer;
   w,ds,ms,ss:string;
   dv,mv,sv:extended;
 begin
   ds:='0';
   ms:='0';
   ss:='0.0';
   w:=trim(s);  {add a stopper}
   sign:=+1;  {default sign}
   n:=0;
   if upcase(w[1]) in ['N','S','E','W','-'] then n:=1
   else if upcase(w[length(w)]) in ['N','S','E','W','-'] then n:=length(w);
   if n>0 then
   begin
     if  (upcase(w[n]) in ['W','S','-']) then sign:=-1;
     delete(w,n,1);
   end;
   n:=0;
   fn:=0;
   w:=w+','; {add a "stopper"}
   while n<length(w) do
   begin
     inc(n);
     if (fn=0) and (w[n] in [' ',',',deg,':'])
     then
     begin
       ds:=copy(w,1,n-1);
       delete(w,1,n);
       n:=0;
       while (length(w)>0) and (w[1] in [' ',deg,minmark,',',':'])
       do delete(w,1,1);
       inc(fn);
     end
     else if  w[n] in [' ',minmark,',','''',':'] then
     begin
       if fn=1 then ms:=copy(w,1,n-1)
       else if fn=2 then ss:=copy(w,1,n-1);
       delete(w,1,n);
       n:=0;
       while (length(w)>0) and (w[1] in [' ',minmark,',','''',':'])
       do delete(w,1,1);
       inc(fn);
     end;
   end;
   result:=convertfield(ds,dv,'degrees');
   if result then result:=convertfield(ms,mv,'minutes');
   if result then result:=convertfield(ss,sv,'seconds');
   if result
   then angle:=sign*(dv+mv/60+sv/3600)
   else angle:=0;
 end;


{********* AngleToStr ************}
 function AngleToStr(angle:extended):string;
 {make string representation  of an angle}
       var
         D:integer;
         M,S:extended;
       begin
         d:=Trunc(angle);
         m:=abs(frac(angle)*60);
         s:=frac(M)*60;
         m:=int(M);
         if s>=59.99 then
         begin
           s:=0;
           m:=m+1;
         end;
         if (angle<0) and (d=0) then
         result:=format('-%3d %2d %5.2f',[d,trunc(m),s])
         else result:=format('%3d %2d %5.2f',[d,trunc(m),s]);
       end;

{************* Distance ****************}
function distance(lat1,lon1,lat2,lon2:extended; Units:integer):extended;
{Distance between to points assuming spherical earth}
 var
   theta:extended;
 begin
   theta := lon1 - lon2;
   result:= Sin(deg2rad(lat1)) * Sin(deg2rad(lat2))
   + Cos(deg2rad(lat1)) * Cos(deg2rad(lat2)) * Cos(deg2rad(theta));
   result := rad2deg(Arccos(result))*60*1.1515; {miles per degree (24872 mile circumference)}
   if units=1 then result := result * 1.609344 {miles to kilometers}
   else if units=2 then result := result * 0.8684; {miles to nautical miles}
 end;

{*************** ApproxEllipticalDistance ************}
function ApproxEllipticalDistance(llat1, llon1, llat2, llon2:extended; units:integer):extended;
var lat1,lat2,lon1,lon2:extended;
    f,g,l:extended;
    sing,cosl,cosf,sinl,sinf,cosg:extended;
    S,C,W,R,H1,H2,D:extended;
begin
   lat1 := DEg2RAd(llat1);
   lon1 := -DEg2RAd(llon1);
   lat2 := DEg2RAd(llat2);
   lon2 := -DEg2RAd(llon2);

   F := (lat1 + lat2) / 2.0;
   G := (lat1 - lat2) / 2.0;
   L := (lon1 - lon2) / 2.0;

   sing := sin(G);
   cosl := cos(L);
   cosf := cos(F);
   sinl := sin(L);
   sinf := sin(F);
   cosg := cos(G);

   S  := sing*sing*cosl*cosl + cosf*cosf*sinl*sinl;
   C  := cosg*cosg*cosl*cosl + sinf*sinf*sinl*sinl;
   W  := arctan2(sqrt(S),sqrt(C));
   R  := sqrt((S*C))/W;
   H1 := (3 * R - 1.0) / (2.0 * C);
   H2 := (3 * R + 1.0) / (2.0 * S);
   D  := 2 * W * ERAD;
   result:=(D * (1 + FLATTENING * H1 * sinf*sinf*cosg*cosg -
                 FLATTENING*H2*cosf*cosf*sing*sing))/1.609344;
   if units=1 then result := result * 1.609344 {miles back to kilometers}
   else if units=2 then result := result * 0.8684; {miles to nautical miles}
end;





{****** SphereBtnClick ************}
procedure TForm1.SphereBtnClick(Sender: TObject);
var
  lat1,lon1,lat2,lon2,dist:extended;
begin
  if not strToAngle(Loc1Lat.text, lat1) then showmessage('Format error in Latitude #1')
  else if not strtoangle(Loc1Lon.text,lon1)then showmessage('Format error in Longitude #1')
    else if not strtoangle(Loc2lat.text,lat2)then showmessage('Format error in Latitude #2')
  else if not strtoangle(Loc2Lon.text,lon2) then showmessage('Format error in Longitude #2')
  else
  begin
    dist:=distance(lat1,lon1,lat2,lon2,unitsgrp.itemindex);
    resultslbl.caption:=format('From Lat %s Long %s to Lat %s, Long %s distance is %6.3f %s',
            [angletostr(lat1),angletostr(lon1),angletostr(lat2), angletostr(lon2),dist,unitslbl[unitsgrp.itemindex]]);
  end;
end;

procedure TForm1.StaticText1Click(Sender: TObject);
begin
   ShellExecute(Handle, 'open', 'http://www.delphiforfun.org/',
  nil, nil, SW_SHOWNORMAL) ;
end;

procedure TForm1.EllispeBtnClick(Sender: TObject);
 {http://www.codeguru.com/Cpp/Cpp/algorithms/general/article.php/c5115/}
  var
  lat1,lon1,lat2,lon2,dist:extended;
begin
  if not strToAngle(Loc1Lat.text, lat1) then showmessage('Format error in Latitude #1')
  else if not strtoangle(Loc1Lon.text,lon1)then showmessage('Format error in Longitude #1')
    else if not strtoangle(Loc2lat.text,lat2)then showmessage('Format error in Latitude #2')
  else if not strtoangle(Loc2Lon.text,lon2) then showmessage('Format error in Longitude #2')
  else
  begin
    dist:=ApproxEllipticalDistance(lat1,lon1,lat2,lon2,unitsgrp.itemindex);
    resultslbl.caption:=format('From Lat %s Long %s to Lat %s, Long %s distance is %6.3f %s',
            [angletostr(lat1),angletostr(lon1),angletostr(lat2), angletostr(lon2),dist,unitslbl[unitsgrp.itemindex]]);
  end;
end;



end.
